import numpy as np
import wget
import pandas as pd
from random import randint, sample

#获取pfm
def get_pfm(taxonomic_groups=str('plants'),data_local = None):
    if data_local == None:
        if taxonomic_groups=='plants':
            DATA_URL = 'http://jaspar.genereg.net/download/CORE/JASPAR2020_CORE_plants_non-redundant_pfms_jaspar.txt'
            out_fname='./plants.txt'
            wget.download(DATA_URL, out=out_fname)
            pre_pfm = pd.read_csv('./plants.txt',
                          sep='\t',
                          header=None)
            
        elif taxonomic_groups=='fungi':
            DATA_URL = 'http://jaspar.genereg.net/download/CORE/JASPAR2020_CORE_fungi_non-redundant_pfms_jaspar.txt'
            out_fname='./plants.txt'
            wget.download(DATA_URL, out=out_fname)
            pre_pfm = pd.read_csv('./plants.txt',
                          sep='\t',
                          header=None)
        
        elif taxonomic_groups=='vertebrates':
            DATA_URL = 'http://jaspar.genereg.net/download/CORE/JASPAR2020_CORE_vertebrates_non-redundant_pfms_jaspar.txt'
            out_fname='./plants.txt'
            wget.download(DATA_URL, out=out_fname)
            pre_pfm = pd.read_csv('./plants.txt',
                          sep='\t',
                          header=None)
        
        elif taxonomic_groups=='insects':
            DATA_URL = 'http://jaspar.genereg.net/download/CORE/JASPAR2020_CORE_insects_non-redundant_pfms_jaspar.txt'
            out_fname='./plants.txt'
            wget.download(DATA_URL, out=out_fname)
            pre_pfm = pd.read_csv('./plants.txt',
                          sep='\t',
                          header=None)
            
            
    else:
        #接收从JASPAR下载的文件作为输入
        pre_pfm = pd.read_csv(str(data_local),
                          sep='\t',
                          header=None)
    pfm = []
    for i in range(0,len(pre_pfm),5):
        pfm_sample = []
        for j in range(i+1,i+5):
            str_pfm = pre_pfm.iloc[j,0][4:-1].strip().split()
            int_pfm = [int(k) for k in str_pfm]
            pfm_sample.append(np.array(int_pfm))
        pfm.append(np.array(pfm_sample).astype('float32'))
    #返回记录有多个pfm数组的列表，每个pfm数组的shape均为4*L(L为长度)
    print('There are '+str(len(pfm))+ ' PFMs '+str('!'))
    print('You need to consider whether the number of CNN filters you use is suitable for this initialization method.')
    return pfm


#获取ppm
def get_ppm(pfm_ ):
    ppm = []
    for k in pfm_:
        ppm.append(k)
    for i in range(len(ppm)):
        for j in range(len(ppm[i][0])):
            a = ppm[i][0][j] / (ppm[i][0][j] + ppm[i][1][j] +
                                ppm[i][2][j] + ppm[i][3][j])
            b = ppm[i][1][j] / (ppm[i][0][j] + ppm[i][1][j] +
                                ppm[i][2][j] + ppm[i][3][j])
            c = ppm[i][2][j] / (ppm[i][0][j] + ppm[i][1][j] +
                                ppm[i][2][j] + ppm[i][3][j])
            d = ppm[i][3][j] / (ppm[i][0][j] + ppm[i][1][j] +
                                ppm[i][2][j] + ppm[i][3][j])
            
            ppm[i][0][j] = a
            ppm[i][1][j] = b
            ppm[i][2][j] = c
            ppm[i][3][j] = d
    return ppm

#计算信息熵，越小越好;注意，这里加了1e-5，防止0的影响
def compute_Information_entropy(acgt):
    return -1*np.sum(np.log2(np.array(acgt)+1e-5)* (np.array(acgt)+1e-5))

#获得一组定长的ppm
def get_ppm_L(ppm,L_ = 8):
    ppm_L = []
    num_drop = 0
    num_L = 0
    f = lambda x: compute_Information_entropy(x)
    for i in ppm:
        #如果长度小于指定值
        if len(i[0])<L_:
            num_drop+=1
        #恰好等于全部的保留
        elif  len(i[0])==L_:
            ppm_L.append(i)
            num_L+=1
        #大于的情况，保留信息熵最大的L-mer
        else:
            min_Information_entropy = 2 
            final_number = 0
            for j in range(0,len(i)-L_):
                if  np.sum([f(a) for a in i.T[j:j+L_]])<min_Information_entropy:
                    min_Information_entropy = np.sum([f(a) for a in i.T[j:j+L_]])
                    final_number = j
                else:
                    pass
            ppm_L.append(i[:,final_number:final_number+L_])
    print(str(num_drop)+ ' PFMs with lengths less than the specified length have been screened out.')
    print( 'All '+str(num_L) + ' PFMs of length exactly equal to the specified length are retained.')
    print('For PFMs with lengths greater than the specified length, the segment with the highest information entropy is intercepted.')
    return ppm_L

#获得一组定长的ppm，每个值-0.5，均值近0
def get_ppm_rp25_L(ppm,L_ = 8):
    ppm_L = []
    num_drop = 0
    num_L = 0
    f = lambda x: compute_Information_entropy(x)
    for i in ppm:
        #如果长度小于指定值
        if len(i[0])<L_:
            num_drop+=1
        #恰好等于全部的保留
        elif  len(i[0])==L_:
            ppm_L.append(i-0.25)
            num_L+=1
        #大于的情况，保留信息熵最大的L-mer
        else:
            min_Information_entropy = 2 
            final_number = 0
            for j in range(0,len(i)-L_):
                if  np.sum([f(a) for a in i.T[j:j+L_]])<min_Information_entropy:
                    min_Information_entropy = np.sum([f(a) for a in i.T[j:j+L_]])
                    final_number = j
                else:
                    pass
            ppm_L.append(i[:,final_number:final_number+L_]-0.25)
    print(str(num_drop)+ ' PFMs with lengths less than the specified length have been screened out.')
    print( 'All '+str(num_L) + ' PFMs of length exactly equal to the specified length are retained.')
    print('For PFMs with lengths greater than the specified length, the segment with the highest information entropy is intercepted.')
    return ppm_L
def get_pwm(ppm_L,background_acgt = [0.25,0.25,0.25,0.25]):
    pwm_L = []
    for i in range(len(ppm_L)):
        pwm_L_sample = []
        for j in range(4):
            #这里加1e-3是防止结果中出现0
            #这里的问题在于，正向最多到2，而负向可以到-inf（很大的负数）
            #考虑如何改，或者直接用ppm？
            #这会有一些问题哦——————
            pwm_L_sample.append(list(np.log2((ppm_L[i][j]+1e-2)/background_acgt[j])))
        pwm_L.append(np.array(pwm_L_sample))
    return np.array(pwm_L) 


def filter_initialization_matrix(taxonomic_groups='plants',data_local = None,
                                 filters=64,
                                 L_=8,
                                 pattern='ppm_rp25',
                                 background_acgt=[0.25, 0.25, 0.25, 0.25]):
    print('Note that the base order in the return result matrix is ACGT')
    #只有pwm模式，需要background_acgt
    if pattern == 'ppm_rp25':
        print(
            'You will get the PPM_R25 matrix((the value of each position of the PPM matrix is subtracted by 0.25)) with the specified number and length.'
        )
        pfm = get_pfm(taxonomic_groups,data_local )
        ppm = get_ppm(pfm)
        ppm_L = get_ppm_rp25_L(ppm, L_)
        sample_number = [randint(0, len(ppm_L) - 1) for _ in range(filters)]
        ppm_r25_filters = []
        for i in sample_number:
            ppm_r25_filters.append(ppm_L[i])
        print('You will get the numpy array with shape ',
              str(np.array(ppm_r25_filters).shape))
        print(
            "You can use numpy's swaaxes function to make the dimension transformation suitable for initializing your parameters"
        )
        return np.array(ppm_r25_filters)
    elif pattern == 'ppm':
        print(
            'You will get the PPM matrix with the specified number and length.'
        )
        pfm = get_pfm(taxonomic_groups,data_local )
        ppm = get_ppm(pfm)
        ppm_L = get_ppm_L(ppm, L_)
        sample_number = [randint(0, len(ppm_L)-1) for _ in range(filters)]
        ppm_filters = []
        for i in sample_number:
            ppm_filters.append(ppm_L[i])
        print('You will get the numpy array with shape ',
              str(np.array(ppm_filters).shape))
        print(
            "You can use numpy's swaaxes function to make the dimension transformation suitable for initializing your parameters"
        )
        return np.array(ppm_filters)
    elif pattern == 'pwm':
        print(
            'You will get the PWM matrix of the specified length and the specified number calculated with '
            + str(background_acgt) + ' as the background.',
            'To prevent negative infinity, the value of 1e-2 is added to all positions.'
        )
        pfm = get_pfm(taxonomic_groups,data_local )
        ppm = get_ppm(pfm)
        ppm_L = get_ppm_L(ppm, L_)
        pwm = get_pwm(ppm_L, background_acgt)
        sample_number = [randint(0, len(pwm)-1) for _ in range(filters)]
        pwm_filters = []
        for i in sample_number:
            pwm_filters.append(pwm[i])
        print('You will get the numpy array with shape ',
              str(np.array(pwm_filters).shape))
        print(
            "You can use numpy's swaaxes function to make the dimension transformation suitable for initializing your parameters"
        )
        return np.array(pwm_filters)


import time #内置模块
import pysam
import pyBigWig
import numpy as np
import pandas as pd
from pysam import FastaFile
from scipy.ndimage import gaussian_filter1d


#辅助函数，用于one-hot编码
#用此函数对100万条1000长度的序列编码需要约700秒（GPU02节点）
def one_hot_dna(dna):
    dna_dict={'A':[1.0,0.,0.,0.],'C':[0.,1.0,0.,0.],'G':[0.,0.,1.0,0.],'T':[0.,0.,0.,1.0],'N':[0.,0.,0.,0.]}
    return np.array([dna_dict[k] for k in dna])


#辅助函数，控制获得固定长度开放区序列
def get_new_bed_df(bed_df , seq_len):
    
    diff = np.array(bed_df.end - bed_df.start)-seq_len
    
    new_start_list = list(np.rint(np.array(bed_df.start) + np.floor(diff*0.5)))
    new_end_list = list(np.rint(np.array(bed_df.end) - np.ceil(diff*0.5)))
    
    #到此处为止，未破坏bed文件原始结构
    new_bed3_df = pd.DataFrame({ 'chr':bed_df.chr,'start':new_start_list,'end':new_end_list})
    
    new_bed3_df.start = new_bed3_df.start.astype(int)
    new_bed3_df.end = new_bed3_df.end.astype(int)
    return new_bed3_df


def get_new_bed4_df(bed_df , seq_len):
    
    diff = np.array(bed_df.end - bed_df.start)-seq_len
    
    new_start_list = list(np.rint(np.array(bed_df.start) + np.floor(diff*0.5)))
    new_end_list = list(np.rint(np.array(bed_df.end) - np.ceil(diff*0.5)))
    
    #到此处为止，未破坏bed文件原始结构
    new_bed3_df = pd.DataFrame({ 'chr':bed_df.chr,'start':new_start_list,'end':new_end_list,'classes':bed_df.classes})
    
    new_bed3_df.start = new_bed3_df.start.astype(int)
    new_bed3_df.end = new_bed3_df.end.astype(int)
    return new_bed3_df



#根据控制文件（记录染色体号和长度，tab间隔）过滤样本
def sample_control(bed_df,genome_control_df):
    #Python内置函数locals()，返回记录当前所有局部变量的字典
    #此处用于局部变量的动态创建
    ld = locals() 
    for i in range(len(genome_control_df)):
        bed_chr_ = bed_df[bed_df.chr == genome_control_df.iloc[i,0]]
        bed_chr =  bed_chr_[bed_chr_.end<genome_control_df.iloc[i,1]]
        ld['sample_' + str(genome_control_df.iloc[i,0])] =bed_chr[bed_chr.start>0]
        
    bed_df_all = []
    for j in genome_control_df.chr:
        bed_df_all.append(ld['sample_' + j])
    bed_df_control = pd.concat(bed_df_all,axis=0)
    print(len(bed_df)-len(bed_df_control),'个样本被筛除，因为它们不在control文件中或超越了control文件中规定的界限')
    return bed_df_control


#这里的第二个参数可以给bed_df而非bed原始文件，即只作为辅助函数，不独立
#约91秒可以完成10万1024长度序列的onehot编码并返回
def get_one_hot_seq_list(fasta_path,sample_bed_df):
    print('对',str(len(sample_bed_df)),'条序列进行one-hot编码')
    start_time = time.time()
    fasta_file = pysam.FastaFile( fasta_path) #打开fasta文件,需要对应目录下有fasta文件的fai索引
    seq=[]
    for index,data in sample_bed_df.iterrows():
        seq.append(one_hot_dna(fasta_file.fetch(data.chr,data.start,data.end) ))
    end_time = time.time()
    print('序列one-hot编码完成,该步骤累计耗时：',end_time-start_time,'秒')
    return seq


def get_regression(bw_path,sample_bed_df):
    print('对',str(len(sample_bed_df)),'个样本进行信息提取')
    start_time = time.time()
    bw_file = pyBigWig.open(bw_path)
    bw_regression = []
    for index,data in sample_bed_df.iterrows():
        #bw.stats在范围内取均值
        bw_regression.append(bw_file.stats(str(data.chr),int(data.start),int(data.end))[0])
    bw_file.close()
    end_time = time.time()
    print('信息提取完成,该步骤累计耗时：',end_time-start_time,'秒')
    return bw_regression  

def get_base_regression(bw_path,sample_bed_df):
    print('对',str(len(sample_bed_df)),'个样本进行碱基分辨率信息提取')
    start_time = time.time()
    bw_file = pyBigWig.open(bw_path)
    bw_base_regression = []
    for index,data in sample_bed_df.iterrows():
        #bw.values取范围内每个点的值
        bw_base_regression.append(bw_file.values(str(data.chr),int(data.start),int(data.end)))
    bw_file.close()
    end_time = time.time()
    print('信息提取完成,该步骤累计耗时：',end_time-start_time,'秒')
    return bw_base_regression 

#获取反向互补序列，用于数据增强
def DNA_complement(sequence):
    trantab = str.maketrans('ACGTacgt', 'TGCAtgca')     #翻译表
    com_sequence = sequence[::-1].translate(trantab)     # 反向、转换互补
    return com_sequence

#获取序列而不是one-hot的序列
def get_seq_list(fasta_path,sample_bed_df):
    fasta_file = pysam.FastaFile( fasta_path) #打开fasta文件,需要对应目录下有fasta文件的fai索引
    seq_=[]
    for index,data in sample_bed_df.iterrows():
        seq_.append(fasta_file.fetch(data.chr,data.start,data.end))
    return seq_


#为了针对分类问题，应该给定bed第四列作为类别标签！
def genome_dataset(bed_path,fasta_path,seq_len=1024,genome_size_control=None,dataset_type='regression',bw_path=None,Data_Augmentation=False):
    #注意匹配，chr01或chr1，有没有0，大小写，和fasta、bw文件比较
    
    #注意容错机制：文件头有无均可读(待实现)
    with open(bed_path,'r')as f:
        #第一行第一个数据单位（应该是chrom或者chrxx）
        chek_bed = f.readline().strip().split('\t')[0]
        if chek_bed[:3]!='chr' and chek_bed[:3]!='CHR' and chek_bed[:3]!='Chr':
            raise IOError("bed文件格式不合规范，请检查！\n 注：各列需以tab间隔，无文件头")
        else:
            print('bed文件检查通过')
        
    if dataset_type=='regression' or dataset_type=='base_regression':
        bed_df = pd.read_csv(bed_path,sep='\t',header=None).iloc[:,:3]
        bed_df.columns = ['chr','start','end']
        #得到定长的bed文件
        new_bed_df = get_new_bed_df(bed_df,seq_len)
    elif dataset_type=='classification':
        bed_df = pd.read_csv(bed_path,sep='\t',header=None).iloc[:,:4]
        bed_df.columns = ['chr','start','end','classes']
        new_bed_df = get_new_bed4_df(bed_df,seq_len)
    else:
        raise Exception("请选择正确的模式")
        
        
    
    #样本控制
    if genome_size_control!=None:
        print('使用控制文件，将过滤长度不合文件要求的序列和未在文件中出现的染色体对应的序列')
        genome_control = pd.read_csv(genome_size_control,sep='\t',names=['chr','control'])
        control_bed_df = sample_control(new_bed_df,genome_control)
    else:
        control_bed_df = new_bed_df
        print('您选择不使用控制文件')
    
  
    #获取原始序列的one-hot编码结果
    seq = get_one_hot_seq_list(fasta_path,control_bed_df)

    sample_name = []
    #获取sample_name，原始
    for k in range(len(new_bed_df)):
        sample_name.append(str(new_bed_df.iloc[k,0])+','+str(new_bed_df.iloc[k,1])+','+str(new_bed_df.iloc[k,2]))
        
    if dataset_type=='classification':
        if Data_Augmentation:
            print('通过取反向互补序列进行数据增强,经此步骤，您最终得到的样本量将翻倍')
            seq_ = get_seq_list(fasta_path,control_bed_df)
            reverse_com_seq = list(map(DNA_complement , seq_))
            #reverse_com_seq要在此处进行one-hot编码
            seq_all = seq+list(map(one_hot_dna,reverse_com_seq))
            m_sample_name = []
            for k in range(len(new_bed_df)):
                m_sample_name.append(str(new_bed_df.iloc[k,0])+','+str(new_bed_df.iloc[k,1])+','+str(new_bed_df.iloc[k,2])+','+'Reverse_complementation')
            sample_name_all = sample_name+m_sample_name
            classes_all = list(new_bed_df.classes)+list(new_bed_df.classes)
            dataset_df = pd.DataFrame({'sample':sample_name_all,'seq_one_hot':seq_all,'classes':classes_all})
            return dataset_df
                 
        else:
            print('您选择不使用数据增强')
            dataset_df = pd.DataFrame({'sample':sample_name,'seq_one_hot':seq,'classes':list(new_bed_df.classes)})
            return dataset_df
    
    if dataset_type=='regression':
        if Data_Augmentation:
            print('通过取反向互补序列进行数据增强,经此步骤，您最终得到的样本量将翻倍')
            #map返回迭代器，需要list转换
            #返回反向互补序列
            #这里错了，放进来的不能是seq
            seq_ = get_seq_list(fasta_path,control_bed_df)
            reverse_com_seq = list(map(DNA_complement , seq_))
            #reverse_com_seq要在此处进行one-hot编码
            seq_all = seq+list(map(one_hot_dna,reverse_com_seq))
            bw_regression = get_regression(bw_path,control_bed_df)
            print('获取反向互补序列的信号')
            bw_regression_all = bw_regression+bw_regression
            m_sample_name = []
            for k in range(len(new_bed_df)):
                m_sample_name.append(str(new_bed_df.iloc[k,0])+','+str(new_bed_df.iloc[k,1])+','+str(new_bed_df.iloc[k,2])+','+'Reverse_complementation')
            sample_name_all = sample_name+m_sample_name
            dataset_df = pd.DataFrame({'sample':sample_name_all,'seq_one_hot':seq_all,'target':bw_regression_all})
            return dataset_df
        else:
            print('您选择不使用数据增强')
            bw_regression = get_regression(bw_path,control_bed_df)
            dataset_df = pd.DataFrame({'sample':sample_name,'seq_one_hot':seq,'target':bw_regression})
            return dataset_df
        
        
        
    if dataset_type=='base_regression':
        if Data_Augmentation:
            print('通过取反向互补序列进行数据增强，经此步骤，您最终得到的样本量将翻倍')
            #map返回迭代器，需要list转换
            #返回反向互补序列
            #这里错了，放进来的不能是seq
            seq_ = get_seq_list(fasta_path,control_bed_df)
            reverse_com_seq = list(map(DNA_complement , seq_))
            #reverse_com_seq要在此处进行one-hot编码
            seq_all = seq+list(map(one_hot_dna,reverse_com_seq))
            bw_base_regression = get_base_regression(bw_path,control_bed_df)
            print('获取反向互补序列的信号')
            bw_base_regression_all = bw_base_regression+bw_base_regression
            m_sample_name = []
            for k in range(len(new_bed_df)):
                m_sample_name.append(str(new_bed_df.iloc[k,0])+','+str(new_bed_df.iloc[k,1])+','+str(new_bed_df.iloc[k,2])+','+'Reverse_complementation')
            sample_name_all = sample_name+m_sample_name
            dataset_df = pd.DataFrame({'sample':sample_name_all,'seq_one_hot':seq_all,'target':bw_base_regression_all})
            return dataset_df
            
        else:
            print('您选择不使用数据增强')
            bw_base_regression = get_base_regression(bw_path,control_bed_df)
            dataset_df = pd.DataFrame({'sample':sample_name,'seq_one_hot':seq,'target':bw_base_regression})
            return dataset_df




import setuptools

with open("README.md",'r') as f:
	description_ = f.read()


setuptools.setup(
	name = "zzx_deep_genome", #模块名称
	version = "0.0.2", #版本号
	author = "zxzong", 
	author_email = '274573863@qq.com',
	description = "A module for constructing parameter initialization matrix of convolutional neural network for genomics using PFM matrix information of transcription factors in Jaspar database.",
    long_description = description_,
    long_description_content_type = 'text/markdown',
    url = '',
    packages = setuptools.find_packages(),#自动寻找项目中导入的模块

    #模块相关元信息（描述信息，说明开发语言、操作系统、版权）
    classifiers = [
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent"
    ],
     
    #需要自动安装的依赖 
    install_requires=[
    "numpy",
    "pandas",
    "wget",
    "scipy",
    "pysam",
    "pyBigWig"
    ],
     
    python_requires=">=3.6"

	)


import numpy as np
import pandas as pd

#对一组序列文件，获取其pfm,ppm,pwm；仍需检查是否正确
def one_hot_dna(dna):
    dna_dict={'A':[1.0,0.,0.,0.],'C':[0.,1.0,0.,0.],'G':[0.,0.,1.0,0.],'T':[0.,0.,0.,1.0],'N':[0.,0.,0.,0.]}
    return np.array([dna_dict[j] for j in dna])

#获取pfm或ppm或pwm矩阵
def get_pwm(file_path,background_acgt = [0.25,0.25,0.25,0.25],type_='pfm'):
    seq_all_list = []
    #这样会漏掉最后一个！！！！！
    with open(file_path,'r')as f:
        seq_list = []
        seq_file = f.readlines()
        for i in seq_file:
            if i[0]=='>':
                seq_all_list.append(seq_list)
                seq_list=[]
            elif i[0]=='A'or i[0]=='C' or i[0]== 'G' or i[0]=='T' or i[0]=='a'or i[0]=='c'or i[0]=='g'or i[0]=='t':
                seq_list.append(i.strip())
            else:
                pass
        #最后一个>对应的不能漏了
        seq_all_list.append(seq_list)
            
    pfm_all = []
    ppm_all = []
    pwm_all = []
    for i in seq_all_list[1:]:
        one_hot_seq = []
        for j in i:
            one_hot_seq.append(one_hot_dna(j))
        
        pfm = np.sum(np.array(one_hot_seq),axis=0).T
        ppm = pfm/np.sum(pfm,axis=0)
        pwm = []
        for k in range(4):
            pwm.append(list(np.log2((ppm[k])/background_acgt[k])))
        pfm_all.append(pfm)
        ppm_all.append(ppm)
        pwm_all.append(pwm)
    if type_ == 'pfm':
        return pfm_all
    if type_ == 'ppm':
        return ppm_all
    if type_ == 'pwm':
        return pwm_all



import pyBigWig
import numpy as np
import pandas as pd
from scipy.ndimage import gaussian_filter1d

#已测试通过
#用于对bw文件进行标准化和平滑，输出新bw文件
def bw_scale(in_path,out_path,standard=32 ):
    in_bw = pyBigWig.open(in_path)
    out_bw = pyBigWig.open(out_path,'w')
    out_bw.addHeader([(k,v) for k,v in in_bw.chroms().items()])
    for i in in_bw.chroms():
        scale_ = sum(in_bw.values(i,0,in_bw.chroms()[i]))/in_bw.chroms()[i]
        scale_bw_list = np.array(in_bw.values(i,0,in_bw.chroms()[i]))/scale_
        #sigma设置了高斯核的标准差
        fil_scale_bw_list = gaussian_filter1d(list(scale_bw_list), sigma=standard, truncate=3)
        out_bw.addEntries(i, 0, values = fil_scale_bw_list.astype('float16'), span=1, step=1)
    out_bw.close()
    in_bw.close()